import os
import pygame
pygame.init()

READABLE_GRAPHICS_FORMATS = ["bmp","png","gif","jpg"]
GRAPHICS_DIRECTORY = "graphics"
TILE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"tiles")
PORTRAIT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"portraits")
FONT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"fonts")
POSTER_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"posters")
SPRITE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"sprites")

DEFAULT_FONT = "Nauvoo.ttf"
TILE_SIZE = 64

## Sprite frame data.
SPRITE_SIZE = [96,128] ## Animation frame size
DEFAULT_SPRITE_FRAMES_PER_ROW = 3
DEFAULT_SPRITE_HOW_MANY_ROWS = 4
DEFAULT_SPRITE_STARTING_FRAME = 7
SPRITE_COLORKEY = (32,156,0,255)

class Sprite:
    def __init__(self,**options):
        pass

class Art:
    """Holds the game's art.

    Covers several categories:
    -Posters: Backgrounds, misc. illustrations
    -Portraits: Character faces
    -Icons: Mini portraits
    -Tiles: small images used to draw a landscape
    -Sprites: objects with animation frames and drawing functions
    -Fonts: Text designs.

    Rather than load the same art repeatedly, this class holds all art
    that's available to the program in subdirectories of \graphics.
    The art is then available to other parts of the program.
    """
    def __init__(self,**options):
        self.fonts = {}
        self.fonts["default"] = pygame.font.Font(os.path.join(FONT_DIRECTORY,DEFAULT_FONT),32)

        self.portraits = {}
        self.icons = {}
        image_filenames = [f for f in os.listdir(PORTRAIT_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
        for image_name in image_filenames:
            surface = pygame.image.load(os.path.join(PORTRAIT_DIRECTORY,image_name)).convert()
            surface.set_colorkey(SPRITE_COLORKEY)
            self.portraits[image_name[:-4]] = surface
            self.icons[image_name[:-4]] = pygame.transform.scale(surface,(TILE_SIZE,TILE_SIZE))

        self.posters = {}
        image_filenames = [f for f in os.listdir(POSTER_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
        for image_name in image_filenames:
            surface = pygame.image.load(os.path.join(POSTER_DIRECTORY,image_name)).convert_alpha()
            self.posters[image_name[:-4]] = surface

        self.tiles = {}
        image_filenames = [f for f in os.listdir(TILE_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
        for image_name in image_filenames:
            surface = pygame.image.load(os.path.join(TILE_DIRECTORY,image_name)).convert()
            self.tiles[image_name[:-4]] = surface

        self.sprite_sheets = {}
        image_filenames = [f for f in os.listdir(SPRITE_DIRECTORY) if f[-3:] in READABLE_GRAPHICS_FORMATS]
        for image_name in image_filenames:
            surface = pygame.image.load(os.path.join(SPRITE_DIRECTORY,image_name)).convert()
            surface.set_colorkey(SPRITE_COLORKEY)
            self.sprite_sheets[image_name[:-4]] = surface

        ## Set up Pygame Rect objects for a standard sprite sheet.
        ## How large is each animation frame in pixels?
        framesize = options.get("framesize",SPRITE_SIZE)
        ## Lines between frames?
        frames_padding = options.get("frames_padding",0)
        frames_outer_padding = options.get("frames_outer_padding",0)
        frames_per_row = options.get("frames_per_row",DEFAULT_SPRITE_FRAMES_PER_ROW)
        frames_how_many_rows = options.get("frames_how_many_rows",DEFAULT_SPRITE_HOW_MANY_ROWS)
        image_clip_region = pygame.rect.Rect(0,0,framesize[0],framesize[1])

        self.animation_frames = []
        for row in range(frames_how_many_rows):
            y1 = (row * framesize[1]) + (row * frames_padding) + frames_outer_padding
            height = framesize[1]
            for frame in range(frames_per_row):
                x1 = (frame * framesize[0]) + (frame * frames_padding) + frames_outer_padding
                width = framesize[0]
                ## Drum_Roll()
                self.animation_frames.append( pygame.rect.Rect(x1,y1,width,height) )
        """We now have a list of Rects representing areas of a sprite sheet
        corresponding to animation frames."""

    def Inventory(self):
        """Show all available files."""
        print "Portraits:"
        for key in self.portraits.keys():
            print "  -"+key
        print "Tiles:"
        for key in self.tiles.keys():
            print "  -"+key
        print "Posters:"
        for key in self.posters.keys():
            print "  -"+key
        print "Fonts:"
        for key in self.fonts.keys():
            print "  -"+key
        print "Sprite sheets:"
        for key in self.sprite_sheets.keys():
            print "  -"+key

    def Write(self,text,font="default",color=(255,255,255)):
        """Return rendered text; shortcut for pygame.font.render."""
        try:
            return self.fonts[font].render(text,1,color)
        except:
            raise "Error: Font '"+font+"' not found."


if __name__ == "__main__":
    ## Run a demo.
    screen = pygame.display.set_mode((800,600))
    art = Art()
