#!/usr/bin/python
"""
Gear Tale
by Kris Schnee.

Contents (search for chapter symbols):
~Header~
~Imported Modules~
~Constants~
~Classes~
~Functions~
~Autorun~ (the code that runs when this program is run)
"""



##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2007.1.26"
__license__ = "Public Domain"


##___________________________________
##      ~Imported Modules~
##___________________________________

## Basic modules
import random
import time ## For FPS counter.
import os

## Third-party modules
import pygame ## Graphics/event/etc. library based on SDL.
from pygame.locals import * ## Event handling constants.

## My modules
import ringtale ## Game framework.
import art
import driftwood ## Interface.
import conch ## Sound/music (should be revised)
import loading ## Loading of characters & scenarios.

##___________________________________
##            ~Constants~
##___________________________________

## Video setup
os.environ["SDL_VIDEO_CENTERED"] = "1" ## Center the graphics window.
SCREEN_SIZE = (768,576)
SCREEN_FLAGS = 0 #FULLSCREEN

## Sound setup.
conch.music = True
conch.sound = False

## Graphics
TILE_GRAPHICS_FORMATS = ["png","bmp","jpg","gif"]
TILE_SIZE = 64
MAP_SIZE = (12,9)
TILE_OFFSET_X = -16
TILE_OFFSET_Y = -64

GRAPHICS_DIRECTORY = "graphics"
TILE_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"tiles")
PORTRAIT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"portraits")
PORTRAIT_DEFAULT = "generic.png"
FONT_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"fonts")
POSTER_DIRECTORY = os.path.join(GRAPHICS_DIRECTORY,"posters")

## Gameplay
TIME_TILL_READY = 100

TERRAIN_CHANCE = ["grass","grass","grass","grass","grass","grass","grass","grass","forest","forest","water","stone","mountains"]
TERRAINS = {"grass":{},
            "forest":{"defense":1.5},
            "water":{"passable":False},
            "stone":{"passable":False},
            "mountains":{"attack":1.5},
            "floor":{},
            }

DIRECTIONS = {"north":0,"east":1,"south":2,"west":3}
NORTH=0
EAST=1
SOUTH=2
WEST=3

## Cheats
DEBUG_ZAP = True ## A "kill all enemies" command.

## Character creation
NAMES_M = ["Abel","Axel","Bann","Bligh","Bonnie","Cid","Cinder","Clyde","Culgan","Deke","Eckert","Finn","Gerhard","Gil","Halleck","Hiro","Hrothgar","Ice","Jak","Kay","Kiln","Knux","Laharl","Lark","Leon","Lex","Maks","Noel","Oliver","Oppenheim","Pershing","Quixote","Rain","Salk","Snow","Stein","Tark","Torn","Venter","Wyrdal","Zel"]
NAMES_F = ["Alice","Alys","Amy","Bastet","Belle","Celia","Charmi","Deliah","Diva","Erin","Etna","Flonne","Gina","Hecate","Hestia","Ieanna","Iris","Jinn","Lucca","Katrin","Kas","Kyra","Leia","Ling","Mina","Minx","Nadia","Nayla","Nell","Nephi","Penelope","Pika","Sally","Sara","Silk","Tara","Teeka","Ursula","Vae","Zelda"]
NAMES_N = ["Alliam","Ander","Bering","Cubert","Deke","Dex","Eckert","Elias","Fnar","Gerad","Haskell","Illiad","Ivo","Kay","Kirin","Kris","Miles","Oak","Mephit","Morgan","Pip","Qwerty","Rahal","Rem","Seed","Synx","Taran","Ulf","Vector","Vicks","Wein","Xiao","Yi","Zenny","Zera"]
SEXES = {"m":"m","f":"f","n":"n","h":"h","male":"m","female":"f"}
SEX_CHANCE = {"m":50,"f":100}

##___________________________________
##            ~Classes~
##___________________________________

class CharacterDisplay( driftwood.Widget ):
    """Custom widget: Display character info."""
    def __init__(self,**options):
        options["coords"] = (0,SCREEN_SIZE[1]-100,SCREEN_SIZE[0],100)
        options["background_color"] = (0,192,192,128)
        options["background_gradient"] = (0,64,64,128)
        driftwood.Widget.__init__(self,**options)
        self.character = options.get("character")

    def SetCharacter(self,character):
        self.character = character
        self.dirty = True

    def Redraw(self):
        self.dirty = False

        self.DrawBackground()
##        pygame.draw.rect(self.surface,(0,64,128,128),(100,0,667,99),2)
        if self.character:
            self.surface.blit(self.character.portrait,(0,0))
            self.surface.blit(art.Write(self.character.name+" ("+self.character.team+")"),(105,0))
            self.surface.blit(art.Write("HP: "+str(self.character.HP)+"/"+str(self.character.HP_max)),(105,32))
            self.surface.blit(art.Write("Energy: "+str(self.character.energy)+"/"+str(self.character.energy_max)),(105,64))
            self.surface.blit(art.Write("Time: "+str(self.character.time+TIME_TILL_READY)+"%"),(305,32))




class Person:
    """A character: human or other, NPC or PC."""
    def __init__(self,**options):
        self.world = options.get("world")
        if not self.world:
            raise "Error: Need to pass a World instance to the Person class."
        self.coords = options.get("coords")

        self.team = options.get("team","Hero")
        self.role = options.get("role")
        ## Load standard data on my character type.
        ## This data can be overridden by options passed on char. creation.
        if self.role:
            try:
                info = self.world.cast.get(self.role)
                """Contrived way of avoiding "stickiness" of overwriting
                what's supposed to be read-only data. If I just did
                "info.update(options)" I'd be overwriting it."""
                for key in info:
                    if not key in options:
                        options[key] = info[key]
            except:
                print "Invalid character role: "+self.role

        self.name = options.get("name")
        if not self.name:
            self.name = self.role.title()

        image = options.get("image")
        if not image:
            image = self.role.lower()
        self.sprite = art.sprite_sheets.get(image)
        self.portrait = art.portraits.get(image,art.portraits["default"])
        self.icon = art.icons.get(image,art.icons["default"])
        self.current_frame_number = 7
        self.current_frame = art.animation_frames[self.current_frame_number]

        self.time = options.get("time",random.randint(0,9))
        self.speed = options.get("speed",5)
        self.HP = options.get("HP",10)
        self.HP_max = self.HP
        self.attack = options.get("attack",3)
        self.defense = options.get("defense",0)
        self.move_range = options.get("move_range",3)

        ## Energy: Special move fuel. Starts at 0.
        self.energy_max = options.get("energy",0)
        self.energy = 0

        self.defending = False
        self.facing = options.get("facing",0) ## North, East, South, West
        if type(self.facing) is str:
            self.facing = DIRECTIONS[self.facing]
        self.SetFacing(self.facing)

        self.action_names = []
        self.actions = {}
        actions = info.get("actions")
        for key in actions:
            self.AddAction(actions[key])

    def AddAction(self,action):
        name = action["name"]
        self.actions[name] = action
        self.action_names.append(name)

    def AdvanceTimer(self,time):
        self.time += self.speed * time

    def ResetTimer(self):
        self.time -= TIME_TILL_READY
        self.time += random.randint(0,9)

    def SetCoords(self,coords):
        self.coords = coords

    def TakeDamage(self,amount):
        """Relies on world to notice I'm out of HP."""
        self.HP -= amount
        if amount < 0:
            self.HP = min(self.HP,self.HP_max)

    def GainEnergy(self,amount):
        if amount < 0:
            self.LoseEnergy(amount)
        else:
            self.energy = min(self.energy+amount,self.energy_max)

    def LoseEnergy(self,amount,drain_only_if_enough=True):
        """Can be used to see if a character has enough for some attack,
        or just to drain it. Normally returns whether there was enough."""
        if amount < 0:
            self.GainEnergy(amount)
        else:
                if amount > self.energy:
                    if drain_only_if_enough:
                        return False ## No drain; presumably a move fails.
                    else:
                        self.energy = 0 ## Drain to zero.
                        return False ## FYI, presumably not used in this context.
                else:
                    self.energy -= amount
                    return True ## OK, there was enough.

    def GetReady(self):
        self.defending = False

    def SetFacing(self,direction):
        self.facing = direction
        self.current_frame_number = [1,4,7,11][direction]
        self.current_frame = art.animation_frames[self.current_frame_number]


class Tile:
    def __init__(self,**options):
        self.terrain = options.get("terrain","grass")
        self.explored = options.get("explored",True)
        self.coords = options.get("coords")
        self.highlight = None
        self.contents = None
        self.move_cost = options.get("move_cost",1)
    def Put(self,what):
        self.contents = what
        what.SetCoords(self.coords)
    def Take(self):
        what = self.contents
        self.contents = None
        what.SetCoords(None)
        return what
    def Highlight(self,color=None):
        self.highlight = color


class Animation:
    def __init__(self,**options):
        self.coords = list(options.get("coords",[0,0])) ## Assumes screen coords
        coords_in_tiles = options.get("coords_in_tiles")
        if coords_in_tiles:
            self.coords = [self.coords[0]*TILE_SIZE+10,self.coords[1]*TILE_SIZE]
        self.time = options.get("time",10)
        self.delay = options.get("delay",0) ## Time before displaying
        self.surface = pygame.surface.Surface((100,100)).convert_alpha()
        self.motion = options.get("motion",(0,-4))
        text = options.get("text")
        if text:
            t = art.Write(text)
            self.surface = t
        else:
            i = options.get("image")
            self.surface = art.posters[i]
            
    def Draw(self):
        if self.delay:
            self.delay -= 1
        else:
            screen.blit(self.surface,self.coords)
            self.time -= 1
            self.coords[0] += self.motion[0]
            self.coords[1] += self.motion[1]



class Display:
    """A system for drawing the screen."""
    def __init__(self,world):
        self.world = world
        
        self.tiles = {}
        tile_image_filenames = [f for f in os.listdir(TILE_DIRECTORY) if f[-3:] in TILE_GRAPHICS_FORMATS]
        for image_name in tile_image_filenames:
            surface = pygame.image.load(os.path.join(TILE_DIRECTORY,image_name)).convert()
            self.tiles[image_name[:-4]] = surface

        self.animations = []

        self.lifemeter_small = pygame.surface.Surface((TILE_SIZE,4))
        self.lifemeter_small.fill((0,255,0))

    def AddAnimation(self,**options):
        self.animations.append(Animation(**options))

    def DrawAnimations(self):
        for a in self.animations:
            a.Draw()
        ## Remove those that are expired.
        self.animations = [a for a in self.animations if a.time >= 0]

    def Draw(self):
        for y in range(MAP_SIZE[1]):
            for x in range(MAP_SIZE[0]):
                tile = self.world.map[x][y]
                if tile.explored:
                    img = self.tiles[tile.terrain]
                else:
                    img = self.tiles["unexplored"]
                screen.blit(img,(x*TILE_SIZE,y*TILE_SIZE))

                if tile.highlight:
                    pygame.draw.rect(screen,tile.highlight,(x*TILE_SIZE,y*TILE_SIZE,TILE_SIZE,TILE_SIZE),1)

        for y in range(MAP_SIZE[1]):
            for x in range(MAP_SIZE[0]):
                contents = self.world.map[x][y].contents
                if contents:
                    screen.blit(contents.sprite,(x*TILE_SIZE+TILE_OFFSET_X,y*TILE_SIZE+TILE_OFFSET_Y),contents.current_frame)
                    HP_share = float(contents.HP) / contents.HP_max
                    size = TILE_SIZE * HP_share
                    screen.blit(self.lifemeter_small,(x*TILE_SIZE,(y+1)*TILE_SIZE-5),(0,0,size,4))

    def ClearHighlights(self):
        for y in range(MAP_SIZE[1]):
            for x in range(MAP_SIZE[0]):
                self.world.map[x][y].Highlight(None)


class World:
    """A model battlefield containing terrain and characters."""
    def __init__(self,**options):
        self.map = []
        for x in range(MAP_SIZE[0]):
            self.map.append([])
            for y in range(MAP_SIZE[1]):                
                terrain = TERRAIN_CHANCE[random.randint(0,len(TERRAIN_CHANCE)-1)]
                self.map[-1].append( Tile(coords=(x,y),terrain=terrain) )

        self.people = []

        self.time = 0 ## Global time.

        self.cast = loading.LoadCharacterData() ## Get all possible characters.

        self.active_character = None
        self.command_in_progress = {}
        self.tiles_allowed = []

    def Setup(self,scenario):
        """Prepare a battlefield and characters for combat."""
        self.people = []
        self.time = 0
        self.active_character = None
        self.command_in_progress = {}
        self.tiles_allowed = []

        self.map = []
        for x in range(MAP_SIZE[0]):
            self.map.append([])
            for y in range(MAP_SIZE[1]):                
                terrain = TERRAIN_CHANCE[random.randint(0,len(TERRAIN_CHANCE)-1)]
                self.map[-1].append( Tile(coords=(x,y),terrain=terrain) )

        if scenario == "test":
            ## Two heroes (one a healer) and a mook.
            self.AddPerson(world=self,name="Tomas",coords=(1,2),team="Gearheart",role="Tomas",facing=EAST)
            self.AddPerson(world=self,name="Ada",coords=(2,4),team="Gearheart",role="Ada",facing=EAST)
            self.AddPerson(world=self,name="Mook",coords=(7,3),team="Malleus",role="Thug",facing=WEST)

        elif scenario == "random":
            people = [RandomCharacter(),RandomCharacter(),RandomCharacter()]
            for person in people:
                OK = False
                while not OK:
                    coords = (random.randint(1,3),random.randint(0,5))
                    if not self.map[coords[0]][coords[1]].contents:
                        OK = True
                self.AddPerson(world=self,name=person[0],role=person[1],coords=coords,team="Gearheart",facing=EAST)
            people = [RandomCharacter(),RandomCharacter(),RandomCharacter()]
            for person in people:
                OK = False
                while not OK:
                    coords = (random.randint(8,10),random.randint(3,8))
                    if not self.map[coords[0]][coords[1]].contents:
                        OK = True
                self.AddPerson(name=person[0],role=person[1],coords=coords,team="Malleus",facing=WEST)

        else:
            try:
                terrain, characters, rules = loading.LoadScenario(scenario)
            except:
                raise "Couldn't load scenario \""+scenario+"\"."
            terrain = terrain.replace("\n","")
            self.map = []
            for x in range(MAP_SIZE[0]):
                self.map.append([])
                for y in range(MAP_SIZE[1]):
                    d = {".":"grass","s":"stone","f":"forest","m":"mountains","w":"water","_":"floor"}
                    terrain_here = d[terrain[ y*MAP_SIZE[0]+x ]]
                    self.map[-1].append( Tile(coords=(x,y),terrain=terrain_here) )

            for line in [l for l in characters.split("\n") if l]:
                line = line.replace(", ",",")
                team, role, coords, facing = line.split(" ")
                coords = coords[1:-1].split(",")
                coords = (int(coords[0]),int(coords[1]))
                self.AddPerson(team=team,role=role,coords=coords,facing=DIRECTIONS[facing])

            ## The Rules are not yet used.

    def SetCommand(self,field,value):
        self.command_in_progress[field] = value

    def ResetCommand(self):
        self.command_in_progress = {}
        self.selected_tile = None

    def SetActiveCharacter(self,who):
        self.active_character = who

    def GetTile(self,coords):
        return self.map[coords[0]][coords[1]]

    def FindTileRange(self,**options):
        """Find all tiles that can be reached by some means.

        Note: Assumes all tiles cost 1 movement point to enter.
        Also: I notice that certain filters get applied per round of
        scanning. This would cause certain weird effects with certain
        parameters such as a chain-lightning attack pattern that can
        only go through a chain of occupied squares.
        """
        coords = options.get("coords",(0,0))
        tile_range = options.get("tile_range",3)
        tiles = [ [coords] ]
        for i in range(tile_range):
            tiles_added_this_round = []
            for t in tiles[-1]:
                new_list = [(t[0],t[1]-1),
                            (t[0],t[1]+1),
                            (t[0]-1,t[1]),
                            (t[0]+1,t[1])]
                new_list = [tile for tile in new_list
                            if not tile in tiles_added_this_round
                            and not tile in tiles[-min(len(tiles),2)] ## Prevents duplicates from "doubling back"
                            and tile[0]>=0
                            and tile[0] < MAP_SIZE[0]
                            and tile[1] >=0
                            and tile[1] < MAP_SIZE[1]]

                ## Eliminate occupied tiles by default.
                if not options.get("occupied_ok"):
                    new_list = [tile for tile in new_list if not self.map[tile[0]][tile[1]].contents]
                ## Eliminate empty tiles?
                if not options.get("empty_ok",True):
                    new_list = [tile for tile in new_list if self.map[tile[0]][tile[1]].contents]
                if not options.get("impassable_ok",False):
                    new_list = [tile for tile in new_list if TERRAINS[self.map[tile[0]][tile[1]].terrain].get("passable",True)]
                ## Eliminate those within a minimum ortho distance?
                if options.get("minimum_range"):
                    pass

                tiles_added_this_round += new_list
            tiles.append(tiles_added_this_round)
        tiles_final = []
        for l in tiles:
            tiles_final += l
        if not options.get("origin_ok",True):
            tiles_final.remove(coords)
        return tiles_final

    def AddPerson(self,**options):
        """Put a character into the game, creating them if needed."""
        options["world"] = self
        person = options.get("person")
        if not person:
            person = Person(**options)
        ## Add to the global list of people.
        self.people.append(person)
        ## Place the person on the field.
        try:
            x,y = options.get("coords")
            self.map[x][y].Put(person)
        except:
            raise "Invalid coordinates."
        return person

    def AdvanceTimer(self,amount=1):
        self.time += 1


class Game( ringtale.Framework ):
    def __init__(self,**options):
        ringtale.Framework.__init__(self)
        self.world = World(**options)
        self.display = Display(self.world)
        self.interface = []
        self.messages = []

        self.story_sequence = "" ## Name of a sequence to load.
        story_sequence_beat = 0 ## Position in story.

        self.pc_team = "Gearheart"
        self.game_over = False

        self.PushState("Title")

    def Notify(self,message):
        self.messages.append(message)

    def FindTileClicked(self,screen_coords):
        tile = (screen_coords[0]/TILE_SIZE,screen_coords[1]/TILE_SIZE)
        if tile[0]<0 or tile[0]>=MAP_SIZE[0] or tile[1]<0 or tile[1]>=MAP_SIZE[1]:
            return None ## Never mind; invalid tile location.
        else:
            return tile

    """ Game States """

    def TitleSetup(self):
        self.interface = []
##        self.interface.append(driftwood.Button(owner=self,name="bTest",coords=(100,350,120,50),text="Test"))
        self.interface.append(driftwood.Button(owner=self,name="bCampaign",coords=(324,350,120,50),text="Campaign"))
        self.interface.append(driftwood.Button(owner=self,name="bRumble",coords=(324,425,120,50),text="Rumble"))
        self.interface.append(driftwood.Button(owner=self,name="bQuit",coords=(324,500,120,50),text="Quit"))
        conch.StopMusic()
        conch.PlaySong("citadel")

    def TitleLogic(self):
        for event in pygame.event.get():
            handled = False
            for widget in self.interface:
                handled = widget.HandleEvent(event)
                if handled:
                    break
            if not handled:
                if event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        self.PopState()

        for message in self.messages:
            text = message.get("text")
            if text == "Clicked":
                sender = message.get("sender")
                if sender == "bQuit":
                    self.PopState()
                    return
                elif sender == "bRumble":
                    self.world.Setup("random")
                    self.PushState("BattleIdle")
                elif sender == "bCampaign":
                    self.SetNote("Still Playing",True)
                    self.SetNote("Campaign Status",1)
                    self.PushState("Campaign")
                elif sender == "bTest":
                    self.PopState()
##                    self.PushState("BattleIdle")
                    self.SetNote("Story Sequence","campaign3")
                    self.PushState("Story")
##                    self.world.Setup("test2")
        self.messages = []

    def TitleDraw(self):
        screen.fill((0,0,0))
        screen.blit(art.posters["gear_logo"],(159,25))
        for element in self.interface:
            element.Draw()

    """ Campaign: Run a series of battles and story scenes.
    Note: This release doesn't preserve character data between battles,
    as a proper RPG would do. There's no experience-level system or
    inventory yet anyway. Note also that the campaign is linear."""

    def CampaignSetup(self):
        conch.StopMusic()
        self.interface = []

    def CampaignLogic(self):
        """Begin the next battle or story scene."""
        status = self.GetNote("Campaign Status")
        ## Has the player made it through all previous battles?
        still_playing = self.GetNote("Still Playing",True)
        if still_playing:
            self.SetNote("Campaign Status",status + 1) ## For next time.
            ## Is there a scenario file for this position, or are we done?
            sequence_name = "campaign"+str(status)
            filename = sequence_name+"."+loading.SCENARIO_FILE_EXTENSION
            if os.path.exists(os.path.join(loading.SCENARIO_DIRECTORY,filename)):
                ## Load the next battle and story.
                self.SetNote("Story Sequence",sequence_name)
                self.world.Setup(sequence_name)
                self.PushState("BattleIdle")
                self.PushState("Story")
            else:
                ## We're done! Show an ending scene, then quit to title.
                self.PopState()
                sequence_name = "campaign"+str(status)
                self.SetNote("Story Sequence",sequence_name)
                self.PushState("Story")

        else:
            ## The player lost.
            print "'But the future refused to change...'"
            self.PopState()

    def CampaignDraw(self):
        pass

    """ Story: Show a non-interactive scene. """

    def StorySetup(self):
        """Partly still uses the explicit variables rather than self.notes.
        It could use either system; which is better?"""
        conch.StopMusic()
        self.notes["interface_text"] = []
        self.story_sequence_beat = 0
        story_sequence = self.GetNote("Story Sequence")
        self.story_sequence = loading.LoadStorySequence(story_sequence)
        beat = self.story_sequence[self.story_sequence_beat]
        song_name = beat.get("song")
        if song_name:
            conch.PlaySong(song_name,True)
            self.story_sequence_beat += 1
            beat = self.story_sequence[self.story_sequence_beat]
        text = beat.get("text")
        text = text.replace(" *"," ")
        text = text.replace("* "," ")

        speaker = beat.get("speaker","")
        portrait = art.portraits.get(speaker)
        speaker = speaker.title()

        self.interface = []
        self.interface.append(driftwood.TextBox(coords=(50,140,668,400),text=text))
        if portrait:
            self.interface.append(driftwood.Widget(coords=(0,0,100,100),background_image=portrait))
        if speaker:
            self.interface.append(driftwood.Widget(coords=(0,100,100,40),text=speaker))

    def StoryLogic(self):
        for event in pygame.event.get():
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    conch.StopMusic()
                    self.PopState()
                    self.story_sequence = None
                    return
                else:
                    self.story_sequence_beat += 1
                    if self.story_sequence_beat == len(self.story_sequence):
                        ## We're done.
                        conch.StopMusic() ## Stop _before_ pop.
                        self.PopState()
                        self.story_sequence = None
                        return

            elif event.type == MOUSEBUTTONDOWN:
                self.story_sequence_beat += 1
                if self.story_sequence_beat == len(self.story_sequence):
                    ## We're done.
                    conch.StopMusic() ## Stop _before_ pop.
                    self.PopState()
                    self.story_sequence = None
                    return

                beat = self.story_sequence[self.story_sequence_beat]
                song_name = beat.get("song")
                while song_name:
                    conch.PlaySong(song_name,True)
                    self.story_sequence_beat += 1
                    beat = self.story_sequence[self.story_sequence_beat]
                    song_name = beat.get("song")

                beat = self.story_sequence[self.story_sequence_beat]
                text = beat.get("text")
                speaker = beat.get("speaker","")
                portrait = art.portraits.get(speaker)
                speaker = speaker.title()

                self.interface = []
                self.interface.append(driftwood.TextBox(coords=(50,140,668,400),text=text))
                if portrait:
                    self.interface.append(driftwood.Widget(coords=(0,0,100,100),background_image=portrait))
                if speaker:
                    self.interface.append(driftwood.Widget(coords=(0,100,100,40),text=speaker))

    def StoryDraw(self):
        screen.fill((0,0,64))
        for i in self.interface:
            i.Draw()

    """ BattleIdle: Main state of combat. """

    def BattleIdleSetup(self):
        self.interface = []
        if not self.pc_team:
            ## The player's team is that of the first character who was loaded.
            self.pc_team = self.world.people[0].team
        conch.PlaySong("ambush")

    def BattleIdleLogic(self):
        if self.display.animations:
            self.PushState("BattleAnimate")
            return

        """To solve the characters-being-skipped problem, I'm thinking of
        moving the activation-of-character code to near the top of this
        function, to right here. Time shouldn't advance while somebody's ready."""

        self.world.AdvanceTimer(1)
        ## Add events happening at specific times/intervals?

        ## Check whether battle is over. (Too often a check?)
        if self.game_over:
            conch.StopMusic()
            self.PopState()
            self.game_over = False
            return

        pc_team_alive = False
        enemy_team_alive = False
        ## Need this people_to_check list in case a character is killed.
        people_to_check = [person for person in self.world.people]
        for person in people_to_check:
            if person.HP <= 0:
                self.world.people.remove(person)
                self.world.map[person.coords[0]][person.coords[1]].Take()
            else:
                if person.team == self.pc_team:
                    pc_team_alive = True
                else:
                    enemy_team_alive = True
        if not pc_team_alive:
            print "--- Team "+self.pc_team+" is defeated! ---"
            self.game_over = True
            self.display.AddAnimation(image="defeat",coords=(184,100),time=20,motion=(0,0))
            
        elif not enemy_team_alive:
            print "--- Team "+self.pc_team+" is victorious! ---"
            self.game_over = True
            self.display.AddAnimation(image="victory",coords=(184,100),time=20,motion=(0,0))

        ## Specialized input handling.
        for event in pygame.event.get():
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.PopState()
                    self.SetNote("Still Playing",False)
                    return
                elif event.key == K_z and DEBUG_ZAP: ## Zap all enemies! A debug command.
                    victims = [person for person in self.world.people if person.team != self.pc_team]
                    for victim in victims:
                        victim.TakeDamage(9999)

        for person in self.world.people:
            ## Advance all timers, making sure none is left out by one filling.
            person.AdvanceTimer(1)
        for person in self.world.people:
            if person.time >= TIME_TILL_READY:
                ## Characters whose gauges fill at exactly the same time
                ## may get skipped over for a round.
                if person.team == self.pc_team:
                    self.world.SetActiveCharacter(person)
                    self.PushState("CommandState")
                else:
                    self.world.SetActiveCharacter(person)
                    self.PushState("NPCCommandState")
                    pass ## No NPC logic exists yet.
                person.GetReady()
                person.ResetTimer()
                return

    def BattleIdleDraw(self):
        self.display.Draw()

    def BattleAnimateLogic(self):
        if not self.display.animations:
            self.PopState()

    def BattleAnimateDraw(self):
        self.display.Draw()
        self.display.DrawAnimations()

    def CommandStateSetup(self):
        self.interface = []
        person = self.world.active_character
        tile = self.world.map[person.coords[0]][person.coords[1]]
        tile.Highlight((255,255,255))

    def CommandStateLogic(self):
        """Begin entering a command (for a PC).

        Triggered when a character's time gauge fills."""
        c = self.world.command_in_progress
        if not c.has_key("move"):
            ## A movement command has not yet been entered. Get one.
            self.world.selected_tile = None
            self.PushState("CommandMove")
            return
        elif not c.has_key("action"):
            self.world.selected_tile = None
            self.PushState("CommandAction")
            return

        ## Then clear the command-in-progress and put it on the timeline.
        self.world.ResetCommand()
        self.PopState()

    def CommandStateDraw(self):
        self.display.Draw()
        for element in self.interface:
            element.Draw()

    def CommandMoveSetup(self):
        self.interface = []
        self.interface.append(CharacterDisplay(character=self.world.active_character))
        person = self.world.active_character
        self.world.tiles_allowed = self.world.FindTileRange(
            tile_range=person.move_range,
            coords=self.world.active_character.coords
            )
        for t in self.world.tiles_allowed:
            self.world.map[t[0]][t[1]].Highlight((0,0,255))
        self.world.map[person.coords[0]][person.coords[1]].Highlight((255,255,255))

    def CommandMoveLogic(self):
        if not self.world.selected_tile:
            self.PushState("CommandPickTarget")
            return
        person = self.world.active_character
        from_tile = self.world.map[person.coords[0]][person.coords[1]]
        from_tile.Take()
        dest_tile = self.world.selected_tile
        dest_tile.Put(person)
        ## Change facing direction.
        facing_direction = FindFacingDirection(from_tile.coords,dest_tile.coords)
        person.SetFacing(facing_direction)            

        self.PopState()
        self.world.SetCommand("move",True)
        self.display.ClearHighlights()

    def CommandMoveDraw(self):
        self.display.Draw()
        for element in self.interface:
            element.Draw()

    """CommandPickTarget: Player clicks on one of several highlighted tiles."""

    def CommandPickTargetSetup(self):
        pass

    def CommandPickTargetLogic(self):
        for event in pygame.event.get():
            if event.type == MOUSEBUTTONDOWN:
                tile = self.FindTileClicked(event.pos)
                if tile and tile in self.world.tiles_allowed:
                    self.world.selected_tile = self.world.map[tile[0]][tile[1]]
                    self.PopState()
                    self.display.ClearHighlights()
                    return

    def CommandPickTargetDraw(self):
        self.display.Draw()
        for element in self.interface:
            element.Draw()

    """CommandAction: Player chooses an action."""

    def CommandActionSetup(self):
        """Player picks an action, then a target."""
        self.interface = []
        actor = self.world.active_character
        self.interface.append(CharacterDisplay(character=actor))

        ## A list of actions available.
        actions = actor.action_names
        y = 0
        for action in actions:
            self.interface.append(driftwood.Button(owner=self,name="b"+action,coords=(578,y,200,64),text=action.replace("_"," ").title()))
            y += 64
        self.SetNote("number_of_action_choices",len(self.world.active_character.action_names))

        ## A box describing the action that the cursor is hovering over.
        self.interface.append(driftwood.Widget(coords=(0,0,578,36),alpha=128,text="Pick an action."))
        self.SetNote("hover_choice",None)

        self.display.ClearHighlights()

    def CommandActionLogic(self):
        if self.world.selected_tile:
            ## Command is ready! Execute it.
            tile = self.world.selected_tile
            ## Currently, assume no AoE effects. Just one tile.
            actor = self.world.active_character
            target = tile.contents

            action = self.world.command_in_progress["action"]
            effect = action.get("effect","damage")

            ## So far, attacks can't affect areas.
            if not target:
                print "No target! Canceling action."
                effect = None

            ## Check for "energy" usage.
            energy_cost = action.get("energy_cost")
            if energy_cost:
                cost_paid = actor.LoseEnergy(energy_cost)
                if not cost_paid:
                    effect = None
            else:
                if effect != "defend":
                    ## Charge with energy.
                    actor.GainEnergy(1)

            ## Apply some effect according to what the move actually does.

            if effect == "defend":
                actor.defending = True

            elif effect == "damage": ## The most common effect.
                power = action.get("power")
                defense = actor.defense

                ## Terrain modifiers.
                defense_mod = TERRAINS[tile.terrain].get("defense",1.0)
                defense *= defense_mod
                power_mod = TERRAINS[tile.terrain].get("damage",1.0)
                power *= power_mod

                damage = max(0,power - 5 + random.randint(0,10) - defense)
                if target.defending:
                    damage = damage / 2
                    self.display.AddAnimation(text="Defended!",coords=target.coords,coords_in_tiles=True)

                ## Apply damage.
                damage = int(damage)
                target.TakeDamage(damage)
                self.display.AddAnimation(text=str(damage),coords=target.coords,coords_in_tiles=True,delay=3)

                ## Play listed sound effect, if any.
                sound = action.get("sound")
                if sound:
                    conch.PlaySound(sound)

            elif effect == "heal":
                power = action.get("power")
                damage = -max(0,power - 5 + random.randint(0,10))
                self.display.AddAnimation(text="Healed!",coords=target.coords,coords_in_tiles=True,delay=0)
                self.display.AddAnimation(text=str(damage),coords=target.coords,coords_in_tiles=True,delay=5)
                target.TakeDamage(damage)

            self.PopState()
            self.world.SetCommand("action",True)
            self.display.ClearHighlights()
            return

        elif self.world.command_in_progress.has_key("action"):
            ## Get data on this action.
            action = self.world.command_in_progress["action"]
            action_range = action.get("action_range",1)
            ## Select a target. Show possible targets.
            self.world.tiles_allowed = self.world.FindTileRange(
                tile_range=action_range,
                coords=self.world.active_character.coords,
                occupied_ok=True,empty_ok=True,impassable_ok=True
                )
            for t in self.world.tiles_allowed:
                self.world.map[t[0]][t[1]].Highlight((255,0,0))
            self.PushState("CommandPickTarget")
            return

        ## Haven't picked an action yet.
        actor = self.world.active_character

        ## Which action, if any, is the cursor over?
        mouse_pos = pygame.mouse.get_pos()
        if mouse_pos[0] > 578:
            choice_number = mouse_pos[1] / 64
            if choice_number < self.GetNote("number_of_action_choices"):
                if choice_number != self.GetNote("hover_choice"):
                    self.SetNote("hover_choice",choice_number)
                    action = actor.actions[actor.action_names[choice_number]]
                    text = actor.action_names[choice_number].title()+": "
                    text += action.get("text","")
                    if action.get("energy_cost"):
                        text += " (Energy cost "+str(action["energy_cost"])+")"
                    self.interface[-1].SetText(text.replace("_"," "))

        for event in pygame.event.get():
            handled = False
            for widget in self.interface:
                handled = widget.HandleEvent(event)
                if handled:
                    break
            if not handled:
                pass ## We're waiting for a button press exclusively.

        for message in self.messages:
            text = message.get("text")
            if text == "Clicked":
                sender = message.get("sender")
                self.world.command_in_progress["action"] = self.world.active_character.actions[sender[1:]]
        self.messages = []

    def CommandActionDraw(self):
        self.display.Draw()
        for element in self.interface:
            element.Draw()

    """ NPC Command State """

    def NPCCommandStateSetup(self):
        person = self.world.active_character
        self.interface = []
        self.interface.append(CharacterDisplay(character=person))

    def NPCCommandStateLogic(self):
        """Runs only once per enemy attack."""
        person = self.world.active_character

        ## Pick a random action that I have enough energy to do.
        ok_action = False
        while not ok_action:
            n = random.randint(0,len(person.action_names)-1)
            action = person.actions[person.action_names[n]]
            if action.get("energy_cost",0) <= person.energy:
                ok_action = True

        ## Choose target. Are any heroes in range?
        tiles_allowed = self.world.FindTileRange(
            tile_range=action.get("action_range",1),
            coords=person.coords,
            occupied_ok=True,origin_ok=False,impassable_ok=True)
        for t in tiles_allowed:
            self.world.map[t[0]][t[1]].Highlight((255,0,0))
        self.world.map[person.coords[0]][person.coords[1]].Highlight((255,255,255))

        ## Here I'm breaking my own design pattern for a deliberate delay.
        self.NPCCommandStateDraw()
        pygame.display.update()
        time.sleep(1)

        for t in tiles_allowed:
            tile = self.world.map[t[0]][t[1]]
            effect = action.get("effect")
            if effect == "heal":
                target_team = person.team
            else:
                target_team = self.pc_team
            if tile.contents and tile.contents.team == target_team:
                ## Aha! There's an enemy in range. Attack!

                target = tile.contents
                attack = action.get("power")
                defense = target.defense

                ## Terrain modifiers.
                defense_mod = TERRAINS[tile.terrain].get("defense",1.0)
                defense *= defense_mod
                attack_mod = TERRAINS[tile.terrain].get("attack",1.0)
                attack *= attack_mod

                if attack < 0:
                    damage = min(0,attack - 5 + random.randint(0,10))
                    self.display.AddAnimation(text="Healed!",coords=target.coords,coords_in_tiles=True,delay=3)
                else:
                    damage = max(0,attack - 50 + random.randint(0,10) - defense)
                    if target.defending:
                        damage = damage / 2
                        self.display.AddAnimation(text="Defended!",coords=target.coords,coords_in_tiles=True)
                ## Apply damage.
                damage = int(damage)
                target.TakeDamage(damage)
                self.display.AddAnimation(text=str(damage),coords=target.coords,coords_in_tiles=True,delay=5)
                self.display.ClearHighlights()
                self.PopState()

                ## Use up energy, or charge it up.
                if action.get("effect") != "defend":
                    if action.get("energy_cost"):
                        person.LoseEnergy(action["energy_cost"])
                    else:
                        person.GainEnergy(1)
                
                return ## Done attacking.

        ## Else, pick a target and approach.
        if action.get("effect") == "heal":
            targets = [other for other in self.world.people if other.team == person.team and other.HP < other.HP_max]
        else:
            targets = [other for other in self.world.people if other.team != person.team]
        destination = None
        if targets:
            target = targets[random.randint(0,len(targets)-1)]
            ## Choose destination.
            tiles_allowed = self.world.FindTileRange(
                tile_range=action.get("action_range",1),
                coords=person.coords,
                occupied_ok=False,origin_ok=False,impassable_ok=False)

            distance = 9999
            for t in tiles_allowed:
                self.world.map[t[0]][t[1]].Highlight((0,0,255))
                d = abs(OrthoDistance(t, target.coords)-action.get("action_rage",1))
                if d < distance:
                    ## This gets me closer to my target.
                    distance = d
                    destination = t
        ## Pause to show where I'm considering moving.
        self.NPCCommandStateDraw()
        pygame.display.update()
        time.sleep(1)
        ## If I found a better place to move to:
        if destination:
            from_tile = self.world.map[person.coords[0]][person.coords[1]]
            from_tile.Take()
            dest_tile = self.world.map[destination[0]][destination[1]]
            dest_tile.Put(person)
            facing_direction = FindFacingDirection(from_tile.coords,dest_tile.coords)
            person.SetFacing(facing_direction)
        self.display.ClearHighlights()

        ## A second attempt at attacking, post-movement.
        ## It's sloppy to duplicate this code.

        ## Choose target. Are any heroes in range?
        tiles_allowed = self.world.FindTileRange(
            tile_range=action.get("action_range",1),
            coords=person.coords,
            occupied_ok=True,origin_ok=False,impassable_ok=True)
        for t in tiles_allowed:
            self.world.map[t[0]][t[1]].Highlight((255,0,0))
        self.world.map[person.coords[0]][person.coords[1]].Highlight((255,255,255))

        ## Here I'm breaking my own design pattern for a deliberate delay.
        self.NPCCommandStateDraw()
        pygame.display.update()
        time.sleep(1)

        for t in tiles_allowed:
            tile = self.world.map[t[0]][t[1]]
            effect = action.get("effect")
            if effect == "heal":
                target_team = person.team
            else:
                target_team = self.pc_team
            if tile.contents and tile.contents.team == target_team:
                ## Aha! There's an enemy in range. Attack!

                target = tile.contents
                attack = action.get("power")
                defense = target.defense

                ## Terrain modifiers.
                defense_mod = TERRAINS[tile.terrain].get("defense",1.0)
                defense *= defense_mod
                attack_mod = TERRAINS[tile.terrain].get("attack",1.0)
                attack *= attack_mod

                if attack < 0:
                    damage = min(0,attack - 5 + random.randint(0,10))
                    self.display.AddAnimation(text="Healed!",coords=target.coords,coords_in_tiles=True,delay=3)
                else:
                    damage = max(0,attack - 5 + random.randint(0,10) - defense)
                    if target.defending:
                        damage = damage / 2
                        self.display.AddAnimation(text="Defended!",coords=target.coords,coords_in_tiles=True)
                ## Apply damage.
                damage = int(damage)
                target.TakeDamage(damage)
                self.display.AddAnimation(text=str(damage),coords=target.coords,coords_in_tiles=True,delay=5)
                self.display.ClearHighlights()
                self.PopState()

                ## Use up energy, or charge it up.
                if action.get("effect") != "defend":
                    if action.get("energy_cost"):
                        person.LoseEnergy(action["energy_cost"])
                    else:
                        person.GainEnergy(1)
                
                return ## Done attacking.

        self.display.ClearHighlights()
        self.PopState()

    def NPCCommandStateDraw(self):
        self.display.Draw()
        t = art.Write("Enemy's Turn: "+self.world.active_character.name)
        screen.blit(t,(10,10))
        for element in self.interface:
            element.Draw()

##___________________________________
##            ~Functions~
##___________________________________

def OrthoDistance(p1,p2):
    return abs(p1[0]-p2[0]) + abs(p1[1]-p2[1])

def RandomSex():
    s = random.randint(0,99)
    if s < SEX_CHANCE["m"]:
        return "male"
    elif s < SEX_CHANCE["f"]:
        return "female"
def RandomName(whatkind="n"):
    if whatkind == "m":
        l = NAMES_M + NAMES_N
    elif whatkind == "f":
        l = NAMES_F + NAMES_N
    else:
        l = NAMES_N
    x = random.randint(0,len(l)-1)
    return l[x]
def RandomCharacter():
    sex = RandomSex()
    name = RandomName(sex)
    role = ["Thug","Troll","Shadow"][random.randint(0,2)]
    return [name, role]

def FindFacingDirection(coords,coords2):
    """A quick and dirty method of picking a new facing direction."""
    dx = coords2[0] - coords[0]
    dy = coords2[1] - coords[1]
    if abs(dy) > abs(dx):
        if dy > 0:
            return SOUTH
        else:
            return NORTH
    else:
        if dx > 0:
            return EAST
        else:
            return WEST
        

##__________________________________
##            ~Autorun~
##__________________________________


##conch.LoadSound("18378_inferno_ftlas.wav","railgun")

pygame.init()

pygame.event.set_allowed(None)
pygame.event.set_allowed((KEYDOWN,KEYUP,MOUSEBUTTONDOWN,MOUSEBUTTONUP))
pygame.display.set_caption("Gear Tale")
icon = pygame.image.load(os.path.join(GRAPHICS_DIRECTORY,"gear32.png")).convert_alpha()
pygame.display.set_icon(icon)
del icon
screen = pygame.display.set_mode((SCREEN_SIZE),SCREEN_FLAGS)

driftwood.DEFAULT_PEN = driftwood.Pen(filename="nauvoo.ttf",size=30,color=(255,255,255))

art = art.Art()
## CC-licensed music from jamedo.com; see credits.
conch.LoadSong("01 - Angel's Tear.ogg","angel_tear")
conch.LoadSong("09 - Straight Into Ambush.ogg","ambush")
conch.LoadSong("03 - Citadel.ogg","citadel")
##conch.LoadSong("06 - Red Fields.ogg","red_fields")
g = Game()
g.MainLoop()
conch.StopMusic()
